//
//  Copyright (C) 2001-2004 HorizonLive.com, Inc.  All Rights Reserved.
//  Copyright (C) 2002 Constantin Kaplinsky.  All Rights Reserved.
//  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
// VncViewer.java - the VNC viewer applet.  This class mainly just sets up the
// user interface, leaving it to the VncCanvas to do the actual rendering of
// a VNC desktop.
//

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import netscape.javascript.JSObject;

public class VncViewer extends java.applet.Applet
  implements java.lang.Runnable, WindowListener {

  boolean inAnApplet = true;
  boolean inSeparateFrame = false;

  //
  // main() is called when run as a java program from the command line.
  // It simply runs the applet inside a newly-created frame.
  //

  public static void main(String[] argv) {
    VncViewer v = new VncViewer();
    v.mainArgs = argv;
    v.inAnApplet = false;
    v.inSeparateFrame = true;

    v.init();
    v.start();
  }

  String[] mainArgs;

  RfbProto rfb;
  Thread rfbThread;

  Frame vncFrame;
  Container vncContainer;
  ScrollPane desktopScrollPane;
  GridBagLayout gridbag;
  ButtonPanel buttonPanel;
  Label connStatusLabel;
  AuthPanel authenticator;
  AuthUnixLoginPanel authenticatorUnixLogin;
  VncCanvas vc;
  OptionsFrame options;
  ClipboardFrame clipboard;
  RecordingFrame rec;

  // Control session recording.
  Object recordingSync;
  String sessionFileName;
  boolean recordingActive;
  boolean recordingStatusChanged;
  String cursorUpdatesDef;
  String eightBitColorsDef;

  // Variables read from parameter values.
  String socketFactory;
  String host;
  int port;
  boolean showControls;
  boolean offerRelogin;
  boolean showOfflineDesktop;
  int deferScreenUpdates;
  int deferCursorUpdates;
  int deferUpdateRequests;

  // A variable to be passed on to detect if this applet is signed or not
  //    -- Sundar
  boolean isSigned;


  boolean disableSSL;
  boolean setServerSize;
  boolean GET;
  String CONNECT;
  String urlPrefix;
  String httpsPort;
  boolean forceProxy;
  boolean trustAllVncCerts;

  // Reference to this applet for inter-applet communication.
  public static java.applet.Applet refApplet;

  // For JavaScript interaction.
  JSObject jso = null;

  //
  // init()
  //

  public void init() {

    if (inAnApplet) {
      if (jso == null) {
        try {
          jso = JSObject.getWindow(this);
        } catch(Exception e) {
          String msg = e.getMessage();
          System.out.println("Can't get JSObject: " + msg);
        }
      }

      javascript("cancelTimeout()");
    }

    readParameters();

    refApplet = this;

    if (inSeparateFrame) {
      vncFrame = new Frame("TightVNC");
      if (!inAnApplet) {
	vncFrame.add("Center", this);
      }
      vncContainer = vncFrame;
    } else {
      vncContainer = this;
    }

    recordingSync = new Object();

    options = new OptionsFrame(this);
    clipboard = new ClipboardFrame(this);
    authenticator = new AuthPanel(this);
    authenticatorUnixLogin = new AuthUnixLoginPanel();
    if (RecordingFrame.checkSecurity())
      rec = new RecordingFrame(this);

    sessionFileName = null;
    recordingActive = false;
    recordingStatusChanged = false;
    cursorUpdatesDef = null;
    eightBitColorsDef = null;

    if (inSeparateFrame)
      vncFrame.addWindowListener(this);

    rfbThread = new Thread(this);
    rfbThread.start();
  }

  public void update(Graphics g) {
  }

  //
  // run() - executed by the rfbThread to deal with the RFB socket.
  //

  public void run() {

    gridbag = new GridBagLayout();
    vncContainer.setLayout(gridbag);

    GridBagConstraints gbc = new GridBagConstraints();
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    gbc.anchor = GridBagConstraints.NORTHWEST;

    if (showControls) {
      buttonPanel = new ButtonPanel(this);
      gridbag.setConstraints(buttonPanel, gbc);
      vncContainer.add(buttonPanel);
    }

    try {
      connectAndAuthenticate();
      doProtocolInitialisation();

      vc = new VncCanvas(this);
      gbc.weightx = 1.0;
      gbc.weighty = 1.0;

      if (inSeparateFrame) {

	// Create a panel which itself is resizeable and can hold
	// non-resizeable VncCanvas component at the top left corner.
	Panel canvasPanel = new Panel();
	canvasPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
	canvasPanel.add(vc);

	// Create a ScrollPane which will hold a panel with VncCanvas
	// inside.
	desktopScrollPane = new ScrollPane(ScrollPane.SCROLLBARS_AS_NEEDED);
	gbc.fill = GridBagConstraints.BOTH;
	gridbag.setConstraints(desktopScrollPane, gbc);
	desktopScrollPane.add(canvasPanel);

	// Finally, add our ScrollPane to the Frame window.
	vncFrame.add(desktopScrollPane);
	vncFrame.setTitle(rfb.desktopName);
	vncFrame.pack();
	vc.resizeDesktopFrame();

      } else {

	// Just add the VncCanvas component to the Applet.
	gridbag.setConstraints(vc, gbc);
	add(vc);
	validate();

      }

      if (showControls)
	buttonPanel.enableButtons();

      moveFocusToDesktop();
      processNormalProtocol();

    } catch (NoRouteToHostException e) {
      fatalError("Network error: no route to server: " + host, e);
    } catch (UnknownHostException e) {
      fatalError("Network error: server name unknown: " + host, e);
    } catch (ConnectException e) {
      fatalError("Network error: could not connect to server: " +
		 host + ":" + port, e);
    } catch (EOFException e) {
      if (showOfflineDesktop) {
	e.printStackTrace();
	System.out.println("Network error: remote side closed connection");
	if (vc != null) {
	  vc.enableInput(false);
	}
	if (inSeparateFrame) {
	  vncFrame.setTitle(rfb.desktopName + " [disconnected]");
	}
	if (rfb != null && !rfb.closed())
	  rfb.close();
	if (showControls && buttonPanel != null) {
	  buttonPanel.disableButtonsOnDisconnect();
	  if (inSeparateFrame) {
	    vncFrame.pack();
	  } else {
	    validate();
	  }
	}
      } else {
	fatalError("Network error: remote side closed connection", e);
      }
    } catch (IOException e) {
      String str = e.getMessage();
      if (str != null && str.length() != 0) {
	fatalError("Network Error: " + str, e);
      } else {
	fatalError(e.toString(), e);
      }
    } catch (Exception e) {
      String str = e.getMessage();
      if (str != null && str.length() != 0) {
	fatalError("Error: " + str, e);
      } else {
	fatalError(e.toString(), e);
      }
    }
    
  }


  //
  // Process RFB socket messages.
  // If the rfbThread is being stopped, ignore any exceptions,
  // otherwise rethrow the exception so it can be handled.
  //
 
  void processNormalProtocol() throws Exception {
    while(true) {
      try {
        vc.processNormalProtocol();
      } catch (ArrayIndexOutOfBoundsException e) {
        System.out.println("Out of bounds exception.");
      } catch (Exception e) {
        if (rfbThread == null) {
	  System.out.println("Ignoring RFB socket exceptions" +
			     " because applet is stopping");
          return;
        } else {
	  throw e;
        }
      }
    }
  }


  //
  // Connect to the RFB server and authenticate the user.
  //

  void connectAndAuthenticate() throws Exception
  {
    showConnectionStatus("Initializing...");
    if (inSeparateFrame) {
      vncFrame.pack();
      vncFrame.show();
    } else {
      validate();
    }

    while (!tryAuthenticate()) {
      authenticator.retry();
      authenticatorUnixLogin.retry();
    }
  }


  //
  // Try to connect and authenticate.
  //

  boolean tryAuthenticate() throws Exception
  {
    showConnectionStatus("Connecting to " + host + ", port " + port + "...");

    rfb = new RfbProto(host, port, this);
    showConnectionStatus("Connected to server");

    rfb.readVersionMsg();
    showConnectionStatus("RFB server supports protocol version " +
			 rfb.serverMajor + "." + rfb.serverMinor);

    rfb.writeVersionMsg();
    showConnectionStatus("Using RFB protocol version " +
			 rfb.clientMajor + "." + rfb.clientMinor);

    int secType = rfb.negotiateSecurity();
    int authType = 0;

    // FIXME: Map security types to authentication schemes in RfbProto.
    switch (secType) {
    case RfbProto.SecTypeNone:
      authType = RfbProto.AuthNone;
      break;
    case RfbProto.SecTypeVncAuth:
      authType = RfbProto.AuthVNC;
      break;
    case RfbProto.SecTypeTight:
      showConnectionStatus("Enabling TightVNC protocol extensions");
      rfb.initCapabilities();
      rfb.setupTunneling();
      authType = rfb.negotiateAuthenticationTight();
      break;
    default:
      throw new Exception("Unknown security type " + secType);
    }

    boolean success = false;

    switch (authType) {
    case RfbProto.AuthNone:
      showConnectionStatus("No authentication needed");
      success = true;
      break;
    case RfbProto.AuthVNC:
      showConnectionStatus("Performing standard VNC authentication");
      if (authenticator.isInteractionNecessary()) {
	showAuthPanel(authenticator);
	authenticator.moveFocusToDefaultField();
      }
      success = authenticator.tryAuthenticate(rfb);
      if (authenticator.isInteractionNecessary()) {
	vncContainer.remove(authenticator);
      } else {
	// Don't retry non-interactive authentication.
	if (!success)
	  throw new Exception("VNC authentication failed");
      }
      break;
    case RfbProto.AuthUnixLogin:
      showConnectionStatus("Performing Unix login-style authentication");
      showAuthPanel(authenticatorUnixLogin);
      authenticatorUnixLogin.moveFocusToDefaultField();
      success = authenticatorUnixLogin.tryAuthenticate(rfb);
      vncContainer.remove(authenticatorUnixLogin);
      break;
    default:
      throw new Exception("Unknown authentication scheme " + authType);
    }

    if (!success)
      rfb.close();

    return success;
  }


  //
  // Show a message describing the connection status.
  // To hide the connection status label, use (msg == null).
  //

  void showConnectionStatus(String msg)
  {
    if (msg == null) {
      if (vncContainer.isAncestorOf(connStatusLabel)) {
	vncContainer.remove(connStatusLabel);
      }
      return;
    }

    System.out.println(msg);

    if (connStatusLabel == null) {
      connStatusLabel = new Label("Status: " + msg);
      connStatusLabel.setFont(new Font("Helvetica", Font.PLAIN, 12));
    } else {
      connStatusLabel.setText("Status: " + msg);
    }

    if (!vncContainer.isAncestorOf(connStatusLabel)) {
      GridBagConstraints gbc = new GridBagConstraints();
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbc.fill = GridBagConstraints.HORIZONTAL;
      gbc.anchor = GridBagConstraints.NORTHWEST;
      gbc.weightx = 1.0;
      gbc.weighty = 1.0;
      gbc.insets = new Insets(20, 30, 20, 30);
      gridbag.setConstraints(connStatusLabel, gbc);
      vncContainer.add(connStatusLabel);
    }

    if (inSeparateFrame) {
      vncFrame.pack();
    } else {
      validate();
    }
  }


  //
  // Show an authentication panel.
  //

  void showAuthPanel(Panel authPanel)
  {
    showConnectionStatus(null);

    GridBagConstraints gbc = new GridBagConstraints();
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    gbc.anchor = GridBagConstraints.NORTHWEST;
    gbc.weightx = 1.0;
    gbc.weighty = 1.0;
    gbc.ipadx = 100;
    gbc.ipady = 50;
    gridbag.setConstraints(authPanel, gbc);
    vncContainer.add(authPanel);

    if (inSeparateFrame) {
      vncFrame.pack();
    } else {
      validate();
    }
  }


  //
  // Do the rest of the protocol initialisation.
  //

  void doProtocolInitialisation() throws IOException
  {
    rfb.writeClientInit();
    rfb.readServerInit();

    if (setServerSize) {
      requestResize(getSize().width, getSize().height);
      rfb.framebufferWidth = getSize().width;
      rfb.framebufferHeight = getSize().height;
    }

    System.out.println("Desktop name is " + rfb.desktopName);
    System.out.println("Desktop size is " + rfb.framebufferWidth + " x " +
		       rfb.framebufferHeight);

    setEncodings();

    showConnectionStatus(null);
  }


  //
  // Send current encoding list to the RFB server.
  //

  int[] encodingsSaved;
  int nEncodingsSaved;

  void setEncodings()        { setEncodings(false); }
  void autoSelectEncodings() { setEncodings(true); }

  void setEncodings(boolean autoSelectOnly) {
    if (options == null || rfb == null || !rfb.inNormalProtocol)
      return;

    int preferredEncoding = options.preferredEncoding;
    if (preferredEncoding == -1) {
      long kbitsPerSecond = rfb.kbitsPerSecond();
      if (nEncodingsSaved < 1) {
        // Choose Tight encoding for the very first update.
        System.out.println("Using Tight encoding");
        preferredEncoding = RfbProto.EncodingTight;
      } else if (kbitsPerSecond > 2000 &&
                 encodingsSaved[0] != RfbProto.EncodingHextile) {
        // Switch to Hextile if the connection speed is above 2Mbps.
        System.out.println("Throughput " + kbitsPerSecond +
                           " kbit/s - changing to Hextile encoding");
        preferredEncoding = RfbProto.EncodingHextile;
      } else if (kbitsPerSecond < 1000 &&
                 encodingsSaved[0] != RfbProto.EncodingTight) {
        // Switch to Tight if the connection speed is below 1Mbps.
        System.out.println("Throughput " + kbitsPerSecond +
                           " kbit/s - changing to Tight encoding");
        preferredEncoding = RfbProto.EncodingTight;
      } else {
        // Don't change the encoder.
        if (autoSelectOnly)
          return;
        preferredEncoding = encodingsSaved[0];
      }
    } else {
      // Auto encoder selection is not enabled.
      if (autoSelectOnly)
        return;
    }

    int[] encodings = new int[20];
    int nEncodings = 0;

    encodings[nEncodings++] = preferredEncoding;
    if (options.useCopyRect) {
      encodings[nEncodings++] = RfbProto.EncodingCopyRect;
    }

    if (preferredEncoding != RfbProto.EncodingHextile) {
      encodings[nEncodings++] = RfbProto.EncodingHextile;
    }
    if (preferredEncoding != RfbProto.EncodingTight) {
      encodings[nEncodings++] = RfbProto.EncodingTight;
    }
    if (preferredEncoding != RfbProto.EncodingZlib) {
      encodings[nEncodings++] = RfbProto.EncodingZlib;
    }
    if (preferredEncoding != RfbProto.EncodingCoRRE) {
      encodings[nEncodings++] = RfbProto.EncodingCoRRE;
    }
    if (preferredEncoding != RfbProto.EncodingRRE) {
      encodings[nEncodings++] = RfbProto.EncodingRRE;
    }

    if (options.compressLevel >= 0 && options.compressLevel <= 9) {
      encodings[nEncodings++] =
        RfbProto.EncodingCompressLevel0 + options.compressLevel;
    }
    if (options.jpegQuality >= 0 && options.jpegQuality <= 9) {
      encodings[nEncodings++] =
        RfbProto.EncodingQualityLevel0 + options.jpegQuality;
    }

    if (options.requestCursorUpdates) {
      encodings[nEncodings++] = RfbProto.EncodingXCursor;
      encodings[nEncodings++] = RfbProto.EncodingRichCursor;
      if (!options.ignoreCursorUpdates)
	encodings[nEncodings++] = RfbProto.EncodingPointerPos;
    }

    encodings[nEncodings++] = RfbProto.EncodingLastRect;
    encodings[nEncodings++] = RfbProto.EncodingNewFBSize;

    boolean encodingsWereChanged = false;
    if (nEncodings != nEncodingsSaved) {
      encodingsWereChanged = true;
    } else {
      for (int i = 0; i < nEncodings; i++) {
        if (encodings[i] != encodingsSaved[i]) {
          encodingsWereChanged = true;
          break;
        }
      }
    }

    if (encodingsWereChanged) {
      try {
        rfb.writeSetEncodings(encodings, nEncodings);
        if (vc != null) {
          vc.softCursorFree();
        }
      } catch (Exception e) {
        e.printStackTrace();
      }
      encodingsSaved = encodings;
      nEncodingsSaved = nEncodings;
    }
  }


  //
  // setCutText() - send the given cut text to the RFB server.
  //

  void setCutText(String text) {
    try {
      if (rfb != null && rfb.inNormalProtocol) {
	rfb.writeClientCutText(text);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }


  //
  // Order change in session recording status. To stop recording, pass
  // null in place of the fname argument.
  //

  void setRecordingStatus(String fname) {
    synchronized(recordingSync) {
      sessionFileName = fname;
      recordingStatusChanged = true;
    }
  }

  //
  // Start or stop session recording. Returns true if this method call
  // causes recording of a new session.
  //

  boolean checkRecordingStatus() throws IOException {
    synchronized(recordingSync) {
      if (recordingStatusChanged) {
	recordingStatusChanged = false;
	if (sessionFileName != null) {
	  startRecording();
	  return true;
	} else {
	  stopRecording();
	}
      }
    }
    return false;
  }

  //
  // Start session recording.
  //

  protected void startRecording() throws IOException {
    synchronized(recordingSync) {
      if (!recordingActive) {
	// Save settings to restore them after recording the session.
	cursorUpdatesDef =
	  options.choices[options.cursorUpdatesIndex].getSelectedItem();
	eightBitColorsDef =
	  options.choices[options.eightBitColorsIndex].getSelectedItem();
	// Set options to values suitable for recording.
	options.choices[options.cursorUpdatesIndex].select("Disable");
	options.choices[options.cursorUpdatesIndex].setEnabled(false);
	options.setEncodings();
	options.choices[options.eightBitColorsIndex].select("No");
	options.choices[options.eightBitColorsIndex].setEnabled(false);
	options.setColorFormat();
      } else {
	rfb.closeSession();
      }

      System.out.println("Recording the session in " + sessionFileName);
      rfb.startSession(sessionFileName);
      recordingActive = true;
    }
  }

  //
  // Stop session recording.
  //

  protected void stopRecording() throws IOException {
    synchronized(recordingSync) {
      if (recordingActive) {
	// Restore options.
	options.choices[options.cursorUpdatesIndex].select(cursorUpdatesDef);
	options.choices[options.cursorUpdatesIndex].setEnabled(true);
	options.setEncodings();
	options.choices[options.eightBitColorsIndex].select(eightBitColorsDef);
	options.choices[options.eightBitColorsIndex].setEnabled(true);
	options.setColorFormat();

	rfb.closeSession();
	System.out.println("Session recording stopped.");
      }
      sessionFileName = null;
      recordingActive = false;
    }
  }


  //
  // readParameters() - read parameters from the html source or from the
  // command line.  On the command line, the arguments are just a sequence of
  // param_name/param_value pairs where the names and values correspond to
  // those expected in the html applet tag source.
  //

  public void readParameters() {
    host = readParameter("HOST", !inAnApplet);
    if (host == null) {
      host = getCodeBase().getHost();
      if (host.equals("")) {
	fatalError("HOST parameter not specified");
      }
    }

    String str = readParameter("PORT", true);
    port = Integer.parseInt(str);

    if (inAnApplet) {
      str = readParameter("Open New Window", false);
      if (str != null && str.equalsIgnoreCase("Yes"))
	inSeparateFrame = true;
    }

    // "Show Controls" set to "No" disables button panel.
    showControls = true;
    str = readParameter("Show Controls", false);
    if (str != null && str.equalsIgnoreCase("No"))
      showControls = false;

    // "Offer Relogin" set to "No" disables "Login again" and "Close
    // window" buttons under error messages in applet mode.
    offerRelogin = true;
    str = readParameter("Offer Relogin", false);
    if (str != null && str.equalsIgnoreCase("No"))
      offerRelogin = false;

    // Do we continue showing desktop on remote disconnect?
    showOfflineDesktop = false;
    str = readParameter("Show Offline Desktop", false);
    if (str != null && str.equalsIgnoreCase("Yes"))
      showOfflineDesktop = true;

    // Fine tuning options.
    deferScreenUpdates = readIntParameter("Defer screen updates", 20);
    deferCursorUpdates = readIntParameter("Defer cursor updates", 10);
    deferUpdateRequests = readIntParameter("Defer update requests", 50);

    // SocketFactory.
    socketFactory = readParameter("SocketFactory", false);

    // SSL
    disableSSL = true;
    str = readParameter("DisableSSL", false);
    if (str != null && str.equalsIgnoreCase("No"))
      disableSSL = false;

    httpsPort = readParameter("httpsPort", false);

    // Extra GET, CONNECT string:
    CONNECT = readParameter("CONNECT", false);
    if (CONNECT != null) {
	CONNECT = CONNECT.replaceAll(" ", ":");
    }

    GET = false;
    str = readParameter("GET", false);
    if (str != null && str.equalsIgnoreCase("Yes")) {
      GET = true;
    }
    if (str != null && str.equalsIgnoreCase("1")) {
      GET = true;
    }

    urlPrefix = readParameter("urlPrefix", false);
    if (urlPrefix != null) {
	urlPrefix = urlPrefix.replaceAll("%2F", "/");
	urlPrefix = urlPrefix.replaceAll("%2f", "/");
	urlPrefix = urlPrefix.replaceAll("_2F_", "/");
	if (urlPrefix.indexOf("/") != 0) {
		urlPrefix = "/" + urlPrefix;
	}
    } else {
        urlPrefix = "";
    }
    System.out.println("urlPrefix: " + urlPrefix);

    forceProxy = false;
    str = readParameter("forceProxy", false);
    if (str != null && str.equalsIgnoreCase("Yes")) {
      forceProxy = true;
    }
    trustAllVncCerts = false;
    str = readParameter("trustAllVncCerts", false);
    if (str != null && str.equalsIgnoreCase("Yes")) {
      trustAllVncCerts = true;
    }
    // Reading the parameter that indicates if the Applet is signed or not
    // --Sundar
    isSigned = false;
    str = readParameter("signed", false);
    if (str != null && str.equalsIgnoreCase("Yes")) {
      isSigned = true;
    }

    // SetServerSize
    setServerSize = false;
    str = readParameter("SetServerSize", false);
    if (str != null && str.equalsIgnoreCase("Yes")) {
      setServerSize = true;
    }
  }

  public String readParameter(String name, boolean required) {
    if (inAnApplet) {
      String s = getParameter(name);
      if ((s == null) && required) {
	fatalError(name + " parameter not specified");
      }
      return s;
    }

    for (int i = 0; i < mainArgs.length; i += 2) {
      if (mainArgs[i].equalsIgnoreCase(name)) {
	try {
	  return mainArgs[i+1];
	} catch (Exception e) {
	  if (required) {
	    fatalError(name + " parameter not specified");
	  }
	  return null;
	}
      }
    }
    if (required) {
      fatalError(name + " parameter not specified");
    }
    return null;
  }

  int readIntParameter(String name, int defaultValue) {
    String str = readParameter(name, false);
    int result = defaultValue;
    if (str != null) {
      try {
	result = Integer.parseInt(str);
      } catch (NumberFormatException e) { }
    }
    return result;
  }

  //
  // moveFocusToDesktop() - move keyboard focus either to the
  // VncCanvas or to the AuthPanel.
  //

  void moveFocusToDesktop() {
    if (vncContainer != null) {
      vncContainer.setFocusCycleRoot(true);
      vncContainer.setFocusTraversalKeysEnabled(false);
      if (vc != null && vncContainer.isAncestorOf(vc)) {
	vc.requestFocus();
      } else if (vncContainer.isAncestorOf(authenticator)) {
	authenticator.moveFocusToDefaultField();
      } else if (vncContainer.isAncestorOf(authenticatorUnixLogin)) {
	authenticatorUnixLogin.moveFocusToDefaultField();
      }
    }
  }

  //
  // disconnect() - close connection to server.
  //

  synchronized public void disconnect() {
    System.out.println("Disconnect");

    if (rfb != null && !rfb.closed())
      rfb.close();
    options.dispose();
    clipboard.dispose();
    if (rec != null)
      rec.dispose();

    if (inAnApplet) {
      showMessage("Disconnected");
    } else {
      System.exit(0);
    }
  }

  //
  // fatalError() - print out a fatal error message.
  // FIXME: Do we really need two versions of the fatalError() method?
  //

  synchronized public void fatalError(String str) {
    System.out.println(str);

    if (inAnApplet) {
      // vncContainer null, applet not inited,
      // can not present the error to the user.
      Thread.currentThread().stop();
    } else {
      System.exit(1);
    }
  }

  synchronized public void fatalError(String str, Exception e) {
 
    if (rfb != null && rfb.closed()) {
      // Not necessary to show error message if the error was caused
      // by I/O problems after the rfb.close() method call.
      System.out.println("RFB thread finished");
      return;
    }

    System.out.println(str);
    e.printStackTrace();

    if (rfb != null)
      rfb.close();

    if (inAnApplet) {
      showMessage(str);
    } else {
      System.exit(1);
    }
  }

  //
  // Show message text and optionally "Relogin" and "Close" buttons.
  //

  void showMessage(String msg) {
    vncContainer.removeAll();

    // Make sure the browser is displaying the applet.
    forceSize();

    if (offerRelogin) {
      ReloginPanel reloginPanel = new ReloginPanel(this, msg);
      vncContainer.add(reloginPanel);
      GridBagConstraints gbc = new GridBagConstraints();
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbc.fill = GridBagConstraints.BOTH;
      gbc.anchor = GridBagConstraints.NORTHWEST;
      gbc.weightx = 1.0;
      gbc.weighty = 1.0;
      gbc.insets = new Insets(8, 8, 8, 8);
      gridbag.setConstraints(reloginPanel, gbc);

    } else {
      Label errLabel = new Label(msg, Label.CENTER);
      errLabel.setFont(new Font("Helvetica", Font.PLAIN, 12));
      vncContainer.add(errLabel);
    }

    if (inSeparateFrame) {
      vncFrame.pack();
    } else {
      validate();
    }
  }

  //
  // Stop the applet.
  // Main applet thread will terminate on first exception
  // after seeing that rfbThread has been set to null.
  //

  public void stop() {
    System.out.println("Stopping applet");
    rfbThread = null;
  }

  //
  // This method is called before the applet is destroyed.
  //

  public void destroy() {
    System.out.println("Destroying applet");

    vncContainer.removeAll();
    options.dispose();
    clipboard.dispose();
    if (rec != null)
      rec.dispose();
    if (rfb != null && !rfb.closed())
      rfb.close();
    if (inSeparateFrame)
      vncFrame.dispose();
  }

  //
  // Start/stop receiving mouse events.
  //

  public void enableInput(boolean enable) {
    vc.enableInput(enable);
  }

  //
  // Close application properly on window close event.
  //

  public void windowClosing(WindowEvent evt) {
    System.out.println("Closing window");
    if (rfb != null)
      disconnect();

    vncContainer.hide();

    if (!inAnApplet) {
      System.exit(0);
    }
  }

  //
  // Move the keyboard focus to the password field on window activation.
  //

  public void windowActivated(WindowEvent evt) {
    if (vncFrame.isAncestorOf(authenticator)) {
      authenticator.moveFocusToDefaultField();
    } else if (vncContainer.isAncestorOf(authenticatorUnixLogin)) {
      authenticatorUnixLogin.moveFocusToDefaultField();
    }
  }

  //
  // Ignore window events we're not interested in.
  //

  public void windowDeactivated (WindowEvent evt) {}
  public void windowOpened(WindowEvent evt) {}
  public void windowClosed(WindowEvent evt) {}
  public void windowIconified(WindowEvent evt) {}
  public void windowDeiconified(WindowEvent evt) {}

  //================================================================
  // The methods below were added for HUB application resize support.
  //================================================================

  // Javascript calls this to refresh the screen.
  public void refresh() {
    if (rfb == null)
      return;
    try {
      rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
                                        rfb.framebufferHeight, false);
    }
    catch(IOException e) {
    }
  }

  // Javascript calls this to request a resize from the VNC server.
  public void requestResize(int w, int h) {
    if (rfb == null)
      return;

    if ((w > 4096) || (w < 0)) {
      System.out.println("ERROR in requestResize: Invalid width " +
                         Integer.toString(w));
    }
    if ((h > 4096) || (h < 0)) {
      System.out.println("ERROR in requestResize: Invalid height " +
                         Integer.toString(w));
    }

    try {
      rfb.writeRequestFBSize(w,h);
    }
    catch(IOException e) {
    }
  }

  // Called internally to force the browser to resize the Applet.
  public void forceSize() {
    int w;
    int h;

    if (rfb == null) {
        w = getSize().width;
        h = getSize().height;
    } else {
        w = rfb.framebufferWidth;
        h = rfb.framebufferHeight;
    }

    if (inSeparateFrame) {
        if (vc != null) /* *njk* *hack* check needed when not run as applet */
            vc.setSize(w, h);
        vncFrame.pack();
        refresh();
        return;
    }

    setSize(w, h);

    String ws = Integer.toString(w);
    String hs = Integer.toString(h);
    System.out.println("forceSize(" + ws + "," + hs + ")");
    javascript("forceSize(" + ws + "," + hs + ")");

    setSize(w, h);

    refresh();
  }

  // Called internally to invoke a client action.
  public void clientAction(String action) {
    javascript("clientAction('" + action + "')");
  }

    // Helper function to execute javascript in the browser.
    public void javascript(String s) {
	if (!inAnApplet) {
            return;
        }
        if (jso == null) {
            System.out.println("JSObject is NULL");
            return;
        }
        try {
          jso.eval(s);
          System.out.println("Executed: " + s);
        }
        catch(Exception e) {
          System.out.println("javascript: Can't call: " + s);
        }
    }

    // Called externally to switch between applet view and separate frame.
    public void popout() {

        vncContainer.removeAll();

        if (inSeparateFrame) {
            inSeparateFrame = false;

            vncFrame.dispose();
            vncFrame = null;
            vncContainer = this;
        } else {
            inSeparateFrame = true;

            vncFrame = new Frame("TightVNC");
            vncFrame.setTitle(rfb.desktopName);
            vncFrame.setResizable(true);
            vncFrame.addWindowListener(this);
            vncContainer = vncFrame;
        }

        gridbag = new GridBagLayout();
        vncContainer.setLayout(gridbag);

        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.anchor = GridBagConstraints.NORTHWEST;

        if (showControls) {
            buttonPanel = new ButtonPanel(this);
            gridbag.setConstraints(buttonPanel, gbc);
            vncContainer.add(buttonPanel);
        }

        gbc.weightx = 1.0;
        gbc.weighty = 1.0;
        gbc.fill = GridBagConstraints.BOTH;
        gridbag.setConstraints(vc, gbc);
        vncContainer.add(vc);

        if (inSeparateFrame) {
            vncFrame.pack();
            vncFrame.show();

            vc.addComponentListener(new ComponentAdapter()
                {
                    Timer updateDelay = null;
                    public void componentResized(ComponentEvent e)
                    {
                        final Component c = e.getComponent();
                        if (updateDelay != null)
                            updateDelay.cancel();
                        updateDelay = new Timer();
                        //System.out.println("Scheduled resize.");
                        updateDelay.schedule(new TimerTask()
                            {
                                public void run() {
                                    //System.out.println("RESIZE");
                                    requestResize(c.getWidth(), c.getHeight());
                                    updateDelay = null;
                                }
                            }, 1000);
                    }
                });
        } else {
            validate();
            forceSize();
        }

    }
}

